# pylint:disable=unused-import
from collections import defaultdict
from typing import List, Tuple, Optional, Iterable, Union, Type, TYPE_CHECKING

from cle import SymbolType

from ...sim_variable import SimMemoryVariable
from .. import Analysis, AnalysesHub
from .condition_processor import ConditionProcessor
from .decompilation_options import DecompilationOption

if TYPE_CHECKING:
    from .peephole_optimizations import PeepholeOptimizationStmtBase, PeepholeOptimizationExprBase


class Decompiler(Analysis):
    def __init__(self, func, cfg=None, options=None, optimization_passes=None, sp_tracker_track_memory=True,
                 peephole_optimizations: Optional[Iterable[Union[Type['PeepholeOptimizationStmtBase'],Type['PeepholeOptimizationExprBase']]]]=None,
                 ):
        self.func = func
        self._cfg = cfg
        self._options = options
        self._optimization_passes = optimization_passes
        self._sp_tracker_track_memory = sp_tracker_track_memory
        self._peephole_optimizations = peephole_optimizations

        self.clinic = None  # mostly for debugging purposes
        self.codegen = None

        self._decompile()

    def _decompile(self):

        if self.func.is_simprocedure:
            return

        options_by_class = defaultdict(list)

        if self._options:
            for o, v in self._options:
                options_by_class[o.cls].append((o, v))

        # set global variables
        self._set_global_variables()

        # convert function blocks to AIL blocks
        clinic = self.project.analyses.Clinic(self.func,
                                              kb=self.kb,
                                              optimization_passes=self._optimization_passes,
                                              sp_tracker_track_memory=self._sp_tracker_track_memory,
                                              cfg=self._cfg,
                                              peephole_optimizations=self._peephole_optimizations,
                                              **self.options_to_params(options_by_class['clinic'])
                                              )
        self.clinic = clinic

        if clinic.graph is None:
            # the function is empty
            return

        cond_proc = ConditionProcessor()

        # recover regions
        ri = self.project.analyses.RegionIdentifier(self.func, graph=clinic.graph, cond_proc=cond_proc, kb=self.kb)

        # structure it
        rs = self.project.analyses.RecursiveStructurer(ri.region, cond_proc=cond_proc, kb=self.kb)

        # simplify it
        s = self.project.analyses.RegionSimplifier(rs.result, kb=self.kb)

        codegen = self.project.analyses.StructuredCodeGenerator(self.func, s.result, cfg=self._cfg,
                                                                func_args=clinic.arg_list,
                                                                kb=self.kb,
                                                                variable_kb=clinic.variable_kb)

        self.codegen = codegen

    def _set_global_variables(self):

        global_variables = self.kb.variables['global']
        for symbol in self.project.loader.main_object.symbols:
            if symbol.type == SymbolType.TYPE_OBJECT:
                global_variables.set_variable('global', symbol.rebased_addr, SimMemoryVariable(symbol.rebased_addr, 1,
                                                                                               name=symbol.name))

    @staticmethod
    def options_to_params(options):
        """
        Convert decompilation options to a dict of params.

        :param List[Tuple[DecompilationOption, Any]] options:   The decompilation options.
        :return:                                                A dict of keyword arguments.
        :rtype:                                                 dict
        """

        d = { }
        for option, value in options:
            d[option.param] = value
        return d


AnalysesHub.register_default('Decompiler', Decompiler)



import operator
from .plugin import SimStatePlugin

class SimStateCGC(SimStatePlugin):
    """
    This state plugin keeps track of CGC state.
    """

    # CGC error codes
    EBADF = 1
    EFAULT = 2
    EINVAL = 3
    ENOMEM = 4
    ENOSYS = 5
    EPIPE = 6

    # other CGC constants
    FD_SETSIZE = 1024
    max_allocation = 0x10000000

    #__slots__ = [ 'heap_location', 'max_str_symbolic_bytes' ]

    def __init__(self):
        SimStatePlugin.__init__(self)

        self.allocation_base = 0xb8000000
        self.time = 0

        self.input_size = 0

        self.input_strings = [ ]
        self.output_strings = [ ]

        self.sinkholes = set()

        self.flag_bytes = None

    @SimStatePlugin.memo
    def copy(self, memo): # pylint: disable=unused-argument
        c = super().copy(memo)

        c.allocation_base = self.allocation_base
        c.time = self.time
        c.input_strings = list(self.input_strings)
        c.output_strings = list(self.output_strings)
        c.input_size = self.input_size
        c.sinkholes = set(self.sinkholes)
        c.flag_bytes = self.flag_bytes

        return c

    def peek_input(self):
        if len(self.input_strings) == 0: return None
        return self.input_strings[0]

    def discard_input(self, num_bytes):
        if len(self.input_strings) == 0: return

        self.input_strings[0] = self.input_strings[0][num_bytes:]
        if self.input_strings[0] == b'':
            self.input_strings.pop(0)

    def peek_output(self):
        if len(self.output_strings) == 0: return None
        return self.output_strings[0]

    def discard_output(self, num_bytes):
        if len(self.output_strings) == 0: return

        self.output_strings[0] = self.output_strings[0][num_bytes:]
        if self.output_strings[0] == b'':
            self.output_strings.pop(0)

    def addr_invalid(self, a):
        return not self.state.solver.solution(a != 0, True)

    def _combine(self, others):
        merging_occured = False

        new_allocation_base = max(o.allocation_base for o in others)
        if self.state.solver.symbolic(new_allocation_base):
            raise ValueError("wat")

        concrete_allocation_base = self.state.solver.eval(self.allocation_base)
        concrete_new_allocation_base = self.state.solver.eval(new_allocation_base)

        if concrete_allocation_base != concrete_new_allocation_base:
            self.allocation_base = new_allocation_base
            merging_occured = True

        return merging_occured

    def merge(self, others, merge_conditions, common_ancestor=None): # pylint: disable=unused-argument
        return self._combine(others)

    def widen(self, others):
        return self._combine(others)

### HEAP MANAGEMENT

    def get_max_sinkhole(self, length):
        """
        Find a sinkhole which is large enough to support `length` bytes.

        This uses first-fit. The first sinkhole (ordered in descending order by their address)
        which can hold `length` bytes is chosen. If there are more than `length` bytes in the
        sinkhole, a new sinkhole is created representing the remaining bytes while the old
        sinkhole is removed.
        """

        ordered_sinks = sorted(list(self.sinkholes), key=operator.itemgetter(0), reverse=True)
        max_pair = None
        for addr, sz in ordered_sinks:
            if sz >= length:
                max_pair = (addr, sz)
                break

        if max_pair is None:
            return None

        remaining = max_pair[1] - length
        max_addr = max_pair[0] + remaining
        max_length = remaining

        self.sinkholes.remove(max_pair)

        if remaining:
            self.sinkholes.add((max_pair[0], max_length))

        return max_addr

    def add_sinkhole(self, address, length):
        """
        Add a sinkhole.

        Allow the possibility for the program to reuse the memory represented by the
        address length pair.
        """

        self.sinkholes.add((address, length))


from angr.sim_state import SimState
SimState.register_default('cgc', SimStateCGC)



# pylint:disable=unused-argument
import claripy
import ailment

from .sequence_walker import SequenceWalker
from .region_identifier import MultiNode
from .structurer_nodes import SequenceNode, CodeNode, ConditionNode, SwitchCaseNode, ConditionalBreakNode, \
    BreakNode, LoopNode
from .condition_processor import ConditionProcessor


class EmptyNodeRemover:
    """
    Rewrites a node and its children to remove empty nodes.

    The following optimizations are performed at the same time:
    - Convert if (A) { } else { ... } to if(!A) { ... } else { }
    """
    def __init__(self, node):
        self.root = node

        handlers = {
            SequenceNode: self._handle_Sequence,
            CodeNode: self._handle_Code,
            ConditionNode: self._handle_Condition,
            SwitchCaseNode: self._handle_SwitchCase,
            LoopNode: self._handle_Loop,

            MultiNode: self._handle_Default,
            BreakNode: self._handle_Default,
            ConditionalBreakNode: self._handle_Default,

            ailment.Block: self._handle_Block,
        }
        self._walker = SequenceWalker(handlers=handlers)
        r = self._walker.walk(self.root)
        if r is None:
            self.result = SequenceNode(nodes=[])
        else:
            self.result = r

    #
    # Handlers
    #

    def _handle_Sequence(self, node, **kwargs):

        new_nodes = [ ]
        for node_ in node.nodes:
            new_node = self._walker._handle(node_)
            if new_node is not None:
                if isinstance(new_node, SequenceNode):
                    new_nodes.extend(new_node.nodes)
                else:
                    new_nodes.append(new_node)

        if not new_nodes:
            return None
        return SequenceNode(nodes=new_nodes)

    def _handle_Code(self, node, **kwargs):
        inner_node = self._walker._handle(node.node)
        if inner_node is None:
            return None
        return CodeNode(inner_node, node.reaching_condition)

    def _handle_Condition(self, node, **kwargs):

        true_node = self._walker._handle(node.true_node)
        false_node = self._walker._handle(node.false_node)

        if true_node is None and false_node is None:
            # empty node
            return None
        if true_node is None and false_node is not None:
            # swap them
            return ConditionNode(node.addr,
                                 node.reaching_condition,
                                 ConditionProcessor.simplify_condition(claripy.Not(node.condition)),
                                 false_node,
                                 false_node=None)
        return ConditionNode(node.addr, node.reaching_condition, node.condition, true_node, false_node=false_node)

    def _handle_Loop(self, node, **kwargs):
        new_seq = self._walker._handle(node.sequence_node)

        if new_seq is None:
            return None

        return LoopNode(node.sort,
                        node.condition,
                        new_seq,
                        addr=node.addr
                        )

    def _handle_SwitchCase(self, node, **kwargs):

        new_cases = { }

        for idx, case in node.cases.items():
            new_case = self._walker._handle(case)
            if new_case is not None:
                new_cases[idx] = new_case

        new_default_node = self._walker._handle(node.default_node)

        if not new_cases and new_default_node is None:
            return None

        return SwitchCaseNode(node.switch_expr,
                              new_cases,
                              new_default_node,
                              addr=node.addr
                              )

    @staticmethod
    def _handle_Default(node, **kwargs):
        return node

    @staticmethod
    def _handle_Block(block, **kwargs):
        if not block.statements:
            return None
        return block



import random
from collections import defaultdict

from . import ExplorationTechnique

class StochasticSearch(ExplorationTechnique):
    """
    Stochastic Search.

    Will only keep one path active at a time, any others will be discarded.
    Before each pass through, weights are randomly assigned to each basic block.
    These weights form a probability distribution for determining which state remains after splits.
    When we run out of active paths to step, we start again from the start state.
    """

    def __init__(self, start_state, restart_prob=0.0001):
        """
        :param start_state:  The initial state from which exploration stems.
        :param restart_prob: The probability of randomly restarting the search (default 0.0001).
        """
        super(StochasticSearch, self).__init__()
        self.start_state = start_state
        self.restart_prob = restart_prob
        self._random = random.Random()
        self._random.seed(42)
        self.affinity = defaultdict(self._random.random)

    def step(self, simgr, stash='active', **kwargs):
        simgr = simgr.step(stash=stash, **kwargs)

        if not simgr.stashes[stash] or self._random.random() < self.restart_prob:
            simgr.stashes[stash] = [self.start_state]
            self.affinity.clear()

        if len(simgr.stashes[stash]) >= 2:
            def weighted_pick(states):
                """
                param states: Diverging states.
                """
                assert len(states) >= 2
                total_weight = sum((self.affinity[s.addr] for s in states))
                selected = self._random.uniform(0, total_weight)
                i = 0
                for i, state in enumerate(states):
                    weight = self.affinity[state.addr]
                    if selected < weight:
                        break
                    else:
                        selected -= weight
                picked = states[i]
                return picked

            simgr.stashes[stash] = [weighted_pick(simgr.stashes[stash])]

        return simgr



import angr
from cle.backends.externs.simdata.io_file import io_file_data_for_arch

import logging

l = logging.getLogger(name=__name__)

######################################
# __getdelim
######################################

class __getdelim(angr.SimProcedure):
    # this code is modified from the 'fgets' implementation
    #   to take an arbitrary delimiter
    #   with no max size for concrete data

    # pylint: disable=arguments-differ
    def run(self, line_ptrptr, len_ptr, delim, file_ptr):
        # let's get the memory back for the file we're interested in and find the delimiter
        fd_offset = io_file_data_for_arch(self.state.arch)['fd']
        fd = self.state.mem[file_ptr + fd_offset:].int.resolved
        simfd = self.state.posix.get_fd(fd)
        if simfd is None:
            return -1


        # symbolic delimiters will make this tricky
        if delim.to_claripy().symbolic:
            raise angr.SimProcedureError("I don't know how to handle a symbolic delimiter")

        # case 1: the data is concrete. we should read it a byte at a time since we can't seek for
        # the newline and we don't have any notion of buffering in-memory
        if simfd.read_storage.concrete:
            realloc = angr.SIM_PROCEDURES['libc']['realloc']

            # #dereference the destination buffer
            line_ptr = self.state.memory.load(line_ptrptr,8)
            size = 120
            # im just always going to realloc and restart at size = 120, regardless of if a proper size buffer exists.
            # this doesn't match the exact behavior of get delim, but is the easiest way to ignore symbolic sizes.
            dst = self.inline_call(realloc, line_ptr, size).ret_expr

            count = 0
            while True:
                data, real_size = simfd.read_data(1)
                if self.state.solver.is_true(real_size == 0):
                    break
                self.state.memory.store(dst + count, data)
                count += 1
                if count == size:
                    size = count + size + 1
                    dst = self.inline_call(realloc, dst, size).ret_expr
                if self.state.solver.is_true(data == delim):
                    break

            self.state.memory.store(dst + count, b'\0')
            self.state.memory.store(len_ptr,count)
            self.state.memory.store(line_ptrptr,dst)
            return count


        # case 2: the data is symbolic, the delimiter could be anywhere. Read some maximum number of bytes
        # and add a constraint to assert the delimiter nonsense.
        # caveat: there could also be no delimiter and the file could EOF.
        else:
            # Just a guess as to a good value for a max size
            size = 1024

            data, real_size = simfd.read_data(size-1)
            delim_byte = chr(self.state.solver.eval(delim))

            for i, byte in enumerate(data.chop(8)):
                self.state.solver.add(self.state.solver.If(
                    i+1 != real_size, byte != delim_byte, # if not last byte returned, not newline
                    self.state.solver.Or(            # otherwise one of the following must be true:
                        i+2 == size,                 # - we ran out of space, or
                        simfd.eof(),                 # - the file is at EOF, or
                        byte == delim_byte                # - it is a newline
                    )))

            malloc = angr.SIM_PROCEDURES['libc']['malloc']

            dst = self.inline_call(malloc,real_size)

            self.state.memory.store(dst, data, size=real_size)
            self.state.memory.store(dst+real_size, b'\0')

            self.state.memory.store(len_ptr,real_size)
            self.state.memory.store(line_ptrptr,dst)

            return real_size




from ..java import JavaSimProcedure


class LoadLibrary(JavaSimProcedure):

    __provides__ = (
        ("java.lang.System", "loadLibrary(java.lang.String)"),
    )

    def run(self, lib): # pylint: disable=arguments-differ,unused-argument
        pass




try:
    import sqlalchemy
except ImportError:
    sqlalchemy = None
    raise ImportError("AngrDB relies on SQLAlchemy. Please install SQLAlchemy first by running:\n"
                      "\tpip install sqlalchemy")

from .db import AngrDB



import angr

# pylint: disable=arguments-differ,unused-argument,no-self-use,inconsistent-return-statements

######################################
# pthread_create
# simulates the new thread as an equally viable branch of symbolic execution
######################################
class pthread_create(angr.SimProcedure):
    ADDS_EXITS = True

    # pylint: disable=unused-argument,arguments-differ
    def run(self, thread, attr, start_routine, arg):
        self.call(start_routine, (arg,), 'terminate_thread')
        self.ret(self.state.solver.BVV(0, self.state.arch.bits))

    def terminate_thread(self, thread, attr, start_routine, arg):
        self.exit(0)

    def static_exits(self, blocks):
        # Execute those blocks with a blank state, and then dump the arguments
        blank_state = angr.SimState(project=self.project, mode="fastpath", cle_memory_backer=self.project.loader.memory)

        # Execute each block
        state = blank_state
        for b in blocks:
            irsb = self.project.factory.default_engine.process(state, b, force_addr=b.addr)
            if irsb.successors:
                state = irsb.successors[0]
            else:
                break

        cc = angr.DEFAULT_CC[self.arch.name](self.arch)
        callfunc = cc.arg(state, 2)
        retaddr = state.memory.load(state.regs.sp, self.arch.bytes)

        all_exits = [
            {'address': callfunc, 'jumpkind': 'Ijk_Call', 'namehint': 'thread_entry'},
            {'address': retaddr, 'jumpkind': 'Ijk_Ret', 'namehint': None},
        ]

        return all_exits


######################################
# pthread_cond_signal
# a no-op
######################################

class pthread_cond_signal(angr.SimProcedure):
    def run(self):
        _ = self.arg(0)


######################################
# pthread_mutex_lock
# also a no-op
######################################

class pthread_mutex_lock(angr.SimProcedure):
    def run(self):
        _ = self.arg(0)


######################################
# pthread_mutex_unlock
# also a no-op
######################################

class pthread_mutex_unlock(angr.SimProcedure):
    def run(self):
        _ = self.arg(0)


######################################
# pthread_once
######################################

class pthread_once(angr.SimProcedure):
    def run(self, control, func):
        controlword = self.state.mem[control].char.resolved
        if (controlword & 2).symbolic:
            raise angr.errors.SimProcedureError("Cannot handle symbolic control data for pthread_once")
        if self.state.solver.is_true(controlword & 2 != 0):
            return 0

        controlword |= 2
        self.state.mem[control].char = controlword
        self.call(func, (), 'retsite')

    def retsite(self, control, func):
        return 0

    # TODO: static exits



from typing import List, Union, Set, Tuple
import logging
import operator
import re

from ...engines.light import RegisterOffset
from .constants import DEBUG
from .undefined import Undefined, UNDEFINED
from .unknown_size import UnknownSize, UNKNOWN_SIZE

l = logging.getLogger(name=__name__)


class DataSet:

    __slots__ = ('data', '_bits', '_mask')

    """
    This class represents a set of data.

    Addition and subtraction are performed on the cartesian product of the operands. Duplicate results are removed.
    Data must always include a set.

    :ivar set[Union[int,str,Undefined]] data: The set of data to represent.
    :ivar Union[int,UnknownSize] bits: The size of an element of the set, in number of bits its representation takes.
    """
    maximum_size = 5

    def __init__(self, data: Union[Set[Union[Undefined,RegisterOffset,int]],Undefined,RegisterOffset,int], bits: Union[int,UnknownSize]):
        self.data: Set[Union[Undefined,RegisterOffset,int]] = data if isinstance(data, set) else {data}
        self._bits = bits
        self._mask = (1 << bits) - 1
        self._limit()

    @property
    def bits(self) -> Union[int,UnknownSize]:
        return self._bits

    @property
    def mask(self) -> int:
        return self._mask

    def _limit(self):
        if DEBUG is True:
            # Deterministic limit
            if (len(self.data)) > DataSet.maximum_size:
                data = list(self.data)
                data.sort(key=repr)
                self.data = set(data[:DataSet.maximum_size])
                l.warning('Reached maximum size of DataSet, discarded %s.', str(data[DataSet.maximum_size:]))
        else:
            # Hash dependent implementation
            while len(self.data) > DataSet.maximum_size:
                l.warning('Reached maximum size of DataSet, discarded %s.', str(self.data.pop()))

    def truncate(self, bits):
        if self._bits <= bits:
            return DataSet(self.data, bits)

        def _truncate(datum):
            if isinstance(datum, int):
                mask = (1 << bits) - 1
                return datum & mask
            elif isinstance(datum, str):
                number_of_bytes = bits // 8
                return datum[:number_of_bytes]
            else:
                return datum

        data = { _truncate(d) for d in self.data }
        return DataSet(data, bits)

    def update(self, data):
        if type(data) is DataSet:
            if self.bits != data.bits:
                l.warning('Update with different sizes.')
            self.data.update(data.data)
        else:
            self.data.add(data)
        self._limit()

    def get_first_element(self):
        assert len(self.data) >= 1
        return next(iter(self.data))

    def __len__(self):
        return len(self.data)

    def _un_op(self, op):
        res = set()

        for s in self:
            if type(s) is Undefined:
                res.add(UNDEFINED)
            else:
                try:
                    tmp = op(s)
                    if isinstance(tmp, int):
                        tmp &= self._mask
                    res.add(tmp)
                except TypeError as ex:  # pylint:disable=try-except-raise,unused-variable
                    # l.warning(ex)
                    raise

        return DataSet(res, self._bits)

    def _bin_op(self, other, op):
        if not type(other) is DataSet:
            raise TypeError("_bin_op() only works on another DataSet instance.")

        res = set()

        #if self._bits != other.bits:
        #    l.warning('Binary operation with different sizes.')

        for o in other:
            for s in self:
                if type(o) is Undefined or type(s) is Undefined:
                    res.add(UNDEFINED)
                else:
                    try:
                        tmp = op(s, o)
                        if isinstance(tmp, int):
                            tmp &= self._mask
                        res.add(tmp)
                    except TypeError as ex:  # pylint:disable=try-except-raise,unused-variable
                        # l.warning(ex)
                        raise

        return DataSet(res, self._bits)

    def __add__(self, other):
        return self._bin_op(other, operator.add)

    def __sub__(self, other):
        return self._bin_op(other, operator.sub)

    def __mul__(self, other):
        return self._bin_op(other, operator.mul)

    def __lshift__(self, other):
        return self._bin_op(other, operator.lshift)

    def __rshift__(self, other):
        return self._bin_op(other, operator.rshift)

    def __and__(self, other):
        return self._bin_op(other, operator.and_)

    def __xor__(self, other):
        return self._bin_op(other, operator.xor)

    def __or__(self, other):
        return self._bin_op(other, operator.or_)

    def __neg__(self):
        return self._un_op(operator.neg)

    def __invert__(self):
        return self._un_op(operator.invert)

    def __eq__(self, other):
        if type(other) == DataSet:
            return self.data == other.data and self._bits == other.bits and self._mask == other.mask
        else:
            return False

    def __hash__(self):
        return hash((self._bits, self._mask))

    def __iter__(self):
        return iter(self.data)

    def __str__(self):
        if UNDEFINED in self.data:
            data_string = str(self.data)
        else:
            data_string = str([ _stringify_datum(i) for i in self.data ])
        size = "%d" % self._bits if isinstance(self._bits, int) else self._bits

        return 'DataSet<%s>: %s' % (size, data_string)

def _stringify_datum(datum):
    if isinstance(datum, int): return hex(datum)

    # Shorten long strings if possible
    if isinstance(datum, str) and len(datum) > 50:
        regex = r'(.)\1{10,}'
        matches = list(re.finditer(regex, datum))

        if len(matches) == 0: return datum

        _new_datum = datum[:matches[0].span()[0]]
        for m in matches:
            number_of_occurences = m.span()[1] - m.span()[0]
            _new_datum += "%s...(repeats %d times)" % (m.groups()[0], number_of_occurences)
        return _new_datum

    return datum


def size_of_datasets(datasets: List[DataSet]) -> Tuple[int,UnknownSize]:
    """
    If the size of all datasets are equal, then the result will be of that size; Otherwise, it will be of size UNKNOWN_SIZE.

    :param datasets: The datasets to get the size from.
    :return: The common size of the given datasets.
    """
    if not datasets:
        return UNKNOWN_SIZE
    return (
        datasets[0]._bits if all(map(lambda d: d._bits == datasets[0]._bits, datasets))
        else UNKNOWN_SIZE
    )


def dataset_from_datasets(datasets: List[DataSet]) -> DataSet:
    """
    Instantiate a <DataSet> by "merging" several <DataSet>s.

    :param datasets: The datasets to merge into one.
    :return: A new <DataSet>, containing the data from all given datasets.
    """
    if not datasets:
        return DataSet({UNDEFINED}, UNKNOWN_SIZE)

    size = size_of_datasets(datasets)

    result = DataSet(set(), size)
    for ds in datasets:
        result.update(ds)

    return result




from .base import SimSootValue


class SimSootValue_Local(SimSootValue):

    __slots__ = [ 'id', 'type' ]

    def __init__(self, name, type_):
        super(SimSootValue_Local, self).__init__()
        self.id = name
        self.type = type_

    def __repr__(self):
        return self.id

    @classmethod
    def from_sootvalue(cls, soot_value, state):
        return cls(soot_value.name, soot_value.type)
